package scanner

import (
	"git.urbach.dev/cli/q/src/core"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/fs"
	"git.urbach.dev/cli/q/src/ssa"
	"git.urbach.dev/cli/q/src/token"
)

// scanSignature scans only the function signature without the body.
func scanSignature(file *fs.File, tokens token.List, i int, delimiter token.Kind) (*core.Function, int, error) {
	var (
		groupLevel  = 0
		nameStart   = i
		inputStart  = -1
		inputEnd    = -1
		outputStart = -1
		outputEnd   = -1
	)

	i++

	// Function parameters
	for i < len(tokens) {
		if tokens[i].Kind == token.GroupStart {
			groupLevel++
			i++

			if groupLevel == 1 {
				inputStart = i
			}

			continue
		}

		if tokens[i].Kind == token.GroupEnd {
			groupLevel--

			if groupLevel < 0 {
				return nil, i, errors.New(MissingGroupStart, file, tokens[i].Position)
			}

			if groupLevel == 0 {
				inputEnd = i
				i++
				break
			}

			i++
			continue
		}

		if tokens[i].Kind == token.Invalid {
			return nil, i, errors.New(&InvalidCharacter{Character: tokens[i].String(file.Bytes)}, file, tokens[i].Position)
		}

		if tokens[i].Kind == token.EOF {
			if groupLevel > 0 {
				return nil, i, errors.New(MissingGroupEnd, file, tokens[i].Position)
			}

			return nil, i, errors.New(InvalidFunctionDefinition, file, tokens[i].Position)
		}

		if groupLevel > 0 {
			i++
			continue
		}

		return nil, i, errors.New(InvalidFunctionDefinition, file, tokens[i].Position)
	}

	// Return type
	if i < len(tokens) && tokens[i].Kind == token.ReturnType {
		outputStart = i + 1

		for i < len(tokens) && tokens[i].Kind != delimiter {
			i++
		}

		outputEnd = i
	}

	name := tokens[nameStart].String(file.Bytes)
	function := core.NewFunction(name, file)
	parameters := tokens[inputStart:inputEnd]

	for param := range parameters.Split {
		if len(param) == 0 {
			return nil, i, errors.New(MissingParameter, file, parameters[0].Position)
		}

		if len(param) == 1 {
			return nil, i, errors.New(MissingType, file, param[0].End())
		}

		function.Input = append(function.Input, &ssa.Parameter{
			Name:   param[0].String(file.Bytes),
			Source: ssa.Source(param),
		})
	}

	if outputStart == -1 {
		return function, i, nil
	}

	if tokens[outputStart].Kind == token.GroupStart {
		if tokens[outputEnd-1].Kind == token.GroupEnd {
			outputStart++
			outputEnd--
		} else {
			return nil, i, errors.New(MissingGroupEnd, file, tokens[outputEnd-1].Position)
		}
	}

	outputTokens := tokens[outputStart:outputEnd]

	if len(outputTokens) == 0 {
		return nil, i, errors.New(MissingParameter, file, tokens[outputStart].Position)
	}

	errorPos := token.Position(outputStart)

	for param := range outputTokens.Split {
		if len(param) == 0 {
			return nil, i, errors.New(MissingParameter, file, errorPos)
		}

		if len(param) == 1 {
			function.Output = append(function.Output, &ssa.Parameter{
				Name:   "",
				Source: ssa.Source(param),
			})
		} else {
			function.Output = append(function.Output, &ssa.Parameter{
				Name:   param[0].String(file.Bytes),
				Source: ssa.Source(param),
			})
		}

		errorPos = param[len(param)-1].End() + 1
	}

	return function, i, nil
}